package sun.rmi.transport;

import java.rmi.NoSuchObjectException;
import java.util.Enumeration;
import sun.rmi.runtime.NewThreadAction;
import java.rmi.server.Unreferenced;
import sun.rmi.runtime.Log;
import java.security.AccessController;
import java.security.AccessControlContext;
import java.util.Hashtable;
import java.rmi.dgc.VMID;
import java.util.Vector;
import java.rmi.Remote;
import sun.rmi.server.Dispatcher;
import java.rmi.server.ObjID;

public final class Target
{
    private final ObjID id;
    private final boolean permanent;
    private final WeakRef weakImpl;
    private volatile Dispatcher disp;
    private final Remote stub;
    private final Vector<VMID> refSet;
    private final Hashtable<VMID, SequenceEntry> sequenceTable;
    private final AccessControlContext acc;
    private final ClassLoader ccl;
    private int callCount;
    private boolean removed;
    private volatile Transport exportedTransport;
    private static int nextThreadNum;
    
    public Target(Remote remote, Dispatcher disp, Remote stub, ObjID id, boolean permanent) {
        this.refSet = new Vector();
        this.sequenceTable = new Hashtable(5);
        this.callCount = 0;
        this.removed = false;
        this.exportedTransport = null;
        this.weakImpl = new WeakRef(remote, ObjectTable.reapQueue);
        this.disp = disp;
        this.stub = stub;
        this.id = id;
        this.acc = AccessController.getContext();
        ClassLoader contextClassLoader = Thread.currentThread().getContextClassLoader();
        ClassLoader classLoader = remote.getClass().getClassLoader();
        if (checkLoaderAncestry(contextClassLoader, classLoader)) {
            this.ccl = contextClassLoader;
        }
        else {
            this.ccl = classLoader;
        }
        this.permanent = permanent;
        if (permanent) {
            this.pinImpl();
        }
    }
    
    private static boolean checkLoaderAncestry(ClassLoader parent, ClassLoader classLoader) {
        if (classLoader == null) {
            return true;
        }
        if (parent == null) {
            return false;
        }
        while (parent != null) {
            if (parent == classLoader) {
                return true;
            }
            parent = parent.getParent();
        }
        return false;
    }
    
    public Remote getStub() {
        return this.stub;
    }
    
    ObjectEndpoint getObjectEndpoint() {
        return new ObjectEndpoint(this.id, this.exportedTransport);
    }
    
    WeakRef getWeakImpl() {
        return this.weakImpl;
    }
    
    Dispatcher getDispatcher() {
        return this.disp;
    }
    
    AccessControlContext getAccessControlContext() {
        return this.acc;
    }
    
    ClassLoader getContextClassLoader() {
        return this.ccl;
    }
    
    Remote getImpl() {
        return (Remote)this.weakImpl.get();
    }
    
    boolean isPermanent() {
        return this.permanent;
    }
    
    synchronized void pinImpl() {
        this.weakImpl.pin();
    }
    
    synchronized void unpinImpl() {
        if (!this.permanent && this.refSet.isEmpty()) {
            this.weakImpl.unpin();
        }
    }
    
    void setExportedTransport(Transport exportedTransport) {
        if (this.exportedTransport == null) {
            this.exportedTransport = exportedTransport;
        }
    }
    
    synchronized void referenced(long n, VMID obj) {
        SequenceEntry sequenceEntry = (SequenceEntry)this.sequenceTable.get(obj);
        if (sequenceEntry == null) {
            this.sequenceTable.put(obj, new SequenceEntry(n));
        }
        else if (sequenceEntry.sequenceNum < n) {
            sequenceEntry.update(n);
        }
        else {
            return;
        }
        if (!this.refSet.contains(obj)) {
            this.pinImpl();
            if (this.getImpl() == null) {
                return;
            }
            if (DGCImpl.dgcLog.isLoggable(Log.VERBOSE)) {
                DGCImpl.dgcLog.log(Log.VERBOSE, "add to dirty set: " + obj);
            }
            this.refSet.addElement(obj);
            DGCImpl.getDGCImpl().registerTarget(obj, this);
        }
    }
    
    synchronized void unreferenced(long n, VMID obj, boolean b) {
        SequenceEntry sequenceEntry = (SequenceEntry)this.sequenceTable.get(obj);
        if (sequenceEntry == null || sequenceEntry.sequenceNum > n) {
            return;
        }
        if (b) {
            sequenceEntry.retain(n);
        }
        else if (!sequenceEntry.keep) {
            this.sequenceTable.remove(obj);
        }
        if (DGCImpl.dgcLog.isLoggable(Log.VERBOSE)) {
            DGCImpl.dgcLog.log(Log.VERBOSE, "remove from dirty set: " + obj);
        }
        this.refSetRemove(obj);
        return;
    }
    
    private synchronized void refSetRemove(VMID obj) {
        DGCImpl.getDGCImpl().unregisterTarget(obj, this);
        if (this.refSet.removeElement(obj) && this.refSet.isEmpty()) {
            if (DGCImpl.dgcLog.isLoggable(Log.VERBOSE)) {
                DGCImpl.dgcLog.log(Log.VERBOSE, "reference set is empty: target = " + this);
            }
            Remote impl = this.getImpl();
            if ((impl instanceof Unreferenced)) {
                Unreferenced unreferenced = (Unreferenced)impl;
                ((Thread)AccessController.doPrivileged(new NewThreadAction(()/*  */ -> {
                    Thread.currentThread().setContextClassLoader(this.ccl);
                    AccessController.doPrivileged(()/*  */ -> {
                        var_3_55.unreferenced();
                        return null;
                    }, this.acc);
                    return;
                }, "Unreferenced-" + Target.nextThreadNum++, false, true))).start();
            }
            this.unpinImpl();
        }
    }
    
    synchronized boolean unexport(boolean b) {
        if (b || this.callCount == 0 || this.disp == null) {
            this.disp = null;
            this.unpinImpl();
            DGCImpl dgcImpl = DGCImpl.getDGCImpl();
            Enumeration<VMID> elements = this.refSet.elements();
            while (elements.hasMoreElements()) {
                dgcImpl.unregisterTarget((VMID)elements.nextElement(), this);
            }
            return true;
        }
        return false;
    }
    
    synchronized void markRemoved() {
        if (this.removed) {
            throw new AssertionError();
        }
        this.removed = true;
        if (!this.permanent && this.callCount == 0) {
            ObjectTable.decrementKeepAliveCount();
        }
        if (this.exportedTransport != null) {
            this.exportedTransport.targetUnexported();
        }
        return;
    }
    
    synchronized void incrementCallCount() throws NoSuchObjectException {
        if (this.disp != null) {
            ++this.callCount;
            return;
        }
        throw new NoSuchObjectException("object not accepting new calls");
    }
    
    synchronized void decrementCallCount() {
        int callCount = this.callCount - 1;
        this.callCount = callCount;
        if (callCount < 0) {
            throw new Error("internal error: call count less than zero");
        }
        if (!this.permanent && this.removed && this.callCount == 0) {
            ObjectTable.decrementKeepAliveCount();
        }
        return;
    }
    
    boolean isEmpty() {
        return this.refSet.isEmpty();
    }
    
    public synchronized void vmidDead(VMID vmid) {
        if (DGCImpl.dgcLog.isLoggable(Log.BRIEF)) {
            DGCImpl.dgcLog.log(Log.BRIEF, "removing endpoint " + vmid + " from reference set");
        }
        this.sequenceTable.remove(vmid);
        this.refSetRemove(vmid);
    }
    
    static {
        Target.nextThreadNum = 0;
    }
}
